#include <libpv_config.h>

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#include <libpv.h>

#include <pv_default.h>
#include <pv_types.h>
#include <pv_get.h>
#include <pv_set.h>
#include <pv_bs.h>

/*
 * FIXME: 
 *
 * - different return values in to_device/from_device
 * - make a macro from pv_to_device / pv_from_device
 * - API documentation
 * - all these functions are nearly identical in how they find out about
 *   their pv -> unify. 
 * - document locking policy: the from/to_device functions must be
 *   called with a locked process image (right?)
 */

int pv_to_device(pv_engine_t *engine_anon, void *pv_var)
{
	struct pv_engine	*engine = (struct pv_engine*)engine_anon;	
	struct pv		*pv;
	off_t			pv_var_offset = 0;
	int			finished = 0;

	if (pv_var) {
		pv_var_offset = pv_var - engine->store.map;
		if (pv_var_offset < 0 || pv_var_offset >= engine->store.size)
			return -EINVAL;
	}

	for (goto_head(&engine->pv_list); !finished; goto_next(&engine->pv_list)) {

		if (is_last(engine->pv_list))
			finished=1;

		pv = (struct pv *)list_get_element_data(engine->pv_list);
		if (pv->offset == pv_var_offset) {
			
			if (!pv->device_read)
				return -EINVAL;
					
			pv->device_write(engine_anon, (void*)pv, 0);

			return 0;
		}
	}

	return -EINVAL;
}

int pv_from_device(pv_engine_t *engine_anon, void *pv_var)
{
	struct pv_engine	*engine = (struct pv_engine*)engine_anon;	
	struct pv		*pv;
	off_t			pv_var_offset = 0;
	int			finished = 0;

	if (pv_var) {
		pv_var_offset = pv_var - engine->store.map;
		if (pv_var_offset < 0 || pv_var_offset >= engine->store.size)
			return -EINVAL;
	}

	for (goto_head(&engine->pv_list); !finished; goto_next(&engine->pv_list)) {

		if (is_last(engine->pv_list))
			finished=1;

		pv = (struct pv *)list_get_element_data(engine->pv_list);
		if (pv->offset == pv_var_offset) {
			
			if (!pv->device_write)
				return -EINVAL;
					
			pv->device_read(engine_anon, (void*)pv, 0);

			return 0;
		}
	}

	return -EINVAL;
}


int pv_set_fn_read(pv_engine_t *engine_anon, void *pv_var, int (*device_read)(pv_engine_t *engine, void *pv, unsigned long data))
{
	struct pv_engine	*engine = (struct pv_engine*)engine_anon;	
	struct pv		*pv;
	off_t			pv_var_offset = 0;
	int			finished = 0;

	if (pv_var) {
		pv_var_offset = pv_var - engine->store.map;
		if (pv_var_offset < 0 || pv_var_offset >= engine->store.size)
			return -EINVAL;
	}

	for (goto_head(&engine->pv_list); !finished; goto_next(&engine->pv_list)) {

		if (is_last(engine->pv_list))
			finished=1;

		pv = (struct pv *)list_get_element_data(engine->pv_list);
		if (pv->offset == pv_var_offset) {
			
			pv->device_read = device_read;
			return 0;
		}
	}

	return -EINVAL;
}

int pv_set_fn_write(pv_engine_t *engine_anon, void *pv_var, int (*device_write)(pv_engine_t *engine, void *pv, unsigned long data))
{
	struct pv_engine	*engine = (struct pv_engine*)engine_anon;	
	struct pv		*pv;
	off_t			pv_var_offset = 0;
	int			finished = 0;

	if (pv_var) {
		pv_var_offset = pv_var - engine->store.map;
		if (pv_var_offset < 0 || pv_var_offset >= engine->store.size)
			return -EINVAL;
	}

	for (goto_head(&engine->pv_list); !finished; goto_next(&engine->pv_list)) {

		if (is_last(engine->pv_list))
			finished=1;

		pv = (struct pv *)list_get_element_data(engine->pv_list);
		if (pv->offset == pv_var_offset) {
			
			pv->device_write = device_write;
			return 0;
		}
	}

	return -EINVAL;
}

/*
 * Default read/write functions
 */

/* FIXME: make this also do float stuff */

int pv_device_read_int_file_ascii(pv_engine_t *engine_anon, void *pv_void, unsigned long data)
{
	struct pv_engine *engine = (struct pv_engine*)engine_anon;	
	struct pv *pv = (struct pv *)pv_void;
	char buf[255]; /* FIXME: uggly */
	char *buf_p;
	long ltmp;

	if (pv->device_file == -1)
		return -ENODEV;
	
	lseek(pv->device_file, 0, SEEK_SET);
	for (buf_p=buf; buf_p < buf+sizeof(buf); buf_p++) {
		/* FIXME: make async and check EAGAIN */
		read(pv->device_file, buf_p, 1);
		if (*buf_p == '\n') break;
	}

	if (buf_p > buf+sizeof(buf))
		return -EIO;

	*buf_p = '\0';

	/* FIXME: quick hack... please fix long vs. ulong issues */
	ltmp = strtol(buf, NULL, 0);
	
	switch (pv->type) {
	case PV_TYPE_UINT8:
		pv_set_uint8(engine, pv, (uint8_t)(ltmp)); 
		break;;
	case PV_TYPE_UINT16:
		pv_set_uint16(engine, pv, (uint16_t)(ltmp)); 
		break;;
	case PV_TYPE_UINT32:
		pv_set_uint32(engine, pv, (uint32_t)(ltmp)); 
		break;;
	case PV_TYPE_INT8:
		pv_set_int8(engine, pv, (int8_t)(ltmp)); 
		break;;
	case PV_TYPE_INT16:
		pv_set_int16(engine, pv, (int16_t)(ltmp)); 
		break;;
	case PV_TYPE_INT32:
		pv_set_int32(engine, pv, (int32_t)(ltmp)); 
		break;;
	default:
		printf("unrecognized type\n");
		return -1;
	}
	
	return 0;
}

int pv_device_write_int_file_ascii(pv_engine_t *engine_anon, void *pv_void, unsigned long data)
{
	struct pv_engine *engine = (struct pv_engine*)engine_anon;	
	struct pv *pv = (struct pv *)pv_void;
	char buf[255]; /* FIXME: uggly */
	char *buf_p;
	long ltmp;

	if (pv->device_file == -1)
		return -ENODEV;
	
	lseek(pv->device_file, 0, SEEK_SET);

	/* FIXME: quick hack... please fix long vs. ulong issues */
	
	switch (pv->type) {
	case PV_TYPE_UINT8:
		ltmp = pv_get_uint8(engine, pv); 
		break;;
	case PV_TYPE_UINT16:
		ltmp = pv_get_uint16(engine, pv); 
		break;;
	case PV_TYPE_UINT32:
		ltmp = pv_get_uint32(engine, pv); 
		break;;
	case PV_TYPE_INT8:
		ltmp = pv_get_int8(engine, pv); 
		break;;
	case PV_TYPE_INT16:
		ltmp = pv_get_int16(engine, pv); 
		break;;
	case PV_TYPE_INT32:
		ltmp = pv_get_int32(engine, pv); 
		break;;
	default:
		printf("unrecognized type\n");
		return -1;
	}
	
	snprintf(buf, sizeof(buf), "%li\n", ltmp);
	
	lseek(pv->device_file, 0, SEEK_SET);
	for (buf_p=buf; buf_p < buf+sizeof(buf); buf_p++) {
		/* FIXME: make async and check EAGAIN */
		write(pv->device_file, buf_p, 1);
		if (*buf_p == '\n') break;
	}
	
	return 0;
}

